from typing import Any, Dict
import json
import yaml


class BaseConfig:
    """
    The base configuration class for gallop application
    """
    def __init__(self, **kwargs):
        super().__setattr__("conf_data", kwargs)

    def __getattr__(self, key: str):
        conf_data = self.__getattribute__("conf_data")
        if key in conf_data:
            return conf_data[key]
        else:
            return super().__getattribute__(key)

    def __hasattr__(self, key: str):
        return super().__hasattr__(key)

    def __setattr__(self, key: str, value: Any):
        conf_data = self.__getattribute__("conf_data")
        conf_data[key] = value

    def __getitem__(self, key: str):
        if key not in self.conf_data:
            raise KeyError(f"Config has no key {key}")
        return self.conf_data[key]

    def __setitem__(self, key: str, value: Any):
        self.conf_data[key] = value

    def __contains__(self, key: str):
        return key in self.conf_data

    def __repr__(self) -> str:
        return f"Config({self.conf_data})"

    def __len__(self) -> int:
        return len(self.conf_data)

    def __iter__(self):
        return iter(self.conf_data)

    def items(self):
        return self.conf_data.items()

    def keys(self):
        return self.conf_data.keys()

    # save data to config file
    def to_json(self, path: str):
        with open(path, "w") as f:
            json.dump(self.conf_data, f)

    def to_yaml(self, path: str):
        with open(path, "w") as f:
            yaml.safe_dump(self.conf_data, f)

    # load data from config file
    @classmethod
    def from_json(cls, path: str):
        with open(path, "r") as f:
            conf_data = json.load(f)
        return cls(**conf_data)

    @classmethod
    def from_yaml(cls, path: str):
        with open(path, "r") as f:
            conf_data = yaml.safe_load(f)
        return cls(**conf_data)



from gallop.config import BaseConfig

config = BaseConfig(a=1)


def test_config_set_get():
    assert config.a == 1
    config.a = 2
    assert config["a"] == 2
    config['a'] = 3
    assert config.a == 3


config["b"] = 4
config.c = {"hello": "here"}


def test_another_key():
    assert config.b == 4
    assert len(config) == 3


def test_save_load():
    config.to_json("test/test.json")
    config.to_yaml("test/test.yaml")
    config2 = BaseConfig.from_json("test/test.json")
    config3 = BaseConfig.from_yaml("test/test.yaml")
    assert config2.a == config3.a == 3
    assert config2.b == config3.b == 4
    assert len(config2) == len(config3) == 3


def test_iteration():
    for k in config:
        assert k in "abc"
    for key in config.keys():
        assert key in "abc"
    for key, value in config.items():
        assert key in "abc"
        assert config[key] == value


def test_inheritance():
    class Config(BaseConfig):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.d = 5

    config = Config(a=1)
    assert config.d == 5
    assert config.a == 1
    assert len(config) == 2
    config.to_json("test/test2.json")
    config.to_yaml("test/test2.yaml")
    config2 = Config.from_json("test/test2.json")
    config3 = Config.from_yaml("test/test2.yaml")
    assert config2.a == config3.a == 1
    assert config2.d == config3.d == 5
    assert len(config2) == len(config3) == 2



import setuptools
from configparser import ConfigParser
from pathlib import Path
from pkg_resources import parse_version
assert parse_version(setuptools.__version__) >= parse_version(
    '36.2'), 'setuptools>=36.2 is required'


# note: all settings are in settings.ini; edit there, not here
config = ConfigParser(delimiters=['='])
config.read('settings.ini')
cfg = config['DEFAULT']


cfg_keys = "version description keywords author author_email".split()
expected = (
    cfg_keys
    + "lib_name user branch license status min_python audience language".split()
)

statuses = [
    "1 - Planning",
    "2 - Pre-Alpha",
    "3 - Alpha",
    "4 - Beta",
    "5 - Production/Stable",
    "6 - Mature",
    "7 - Inactive",
]

py_versions = (
    "3.5 3.6 3.7 3.8 3.9 3.10".split()
)

for exp in expected:
    assert exp in cfg, f"missing expected setting: {exp}"
setup_cfg = {k: cfg[k] for k in cfg_keys}

requirements = cfg.get("requirements", "").split()
min_python = cfg["min_python"]
comp_version = py_versions[py_versions.index(min_python):]

with open(f"{cfg['lib_name']}/__init__.py", 'r') as f:
    lines = f.readlines()

with open(f"{cfg['lib_name']}/__init__.py", 'w') as f:
    version = cfg["version"]
    first_line = f'__version__ = "{version}"\n'
    f.write(first_line)
    for line in lines[1:]:
        f.write(line)

setuptools.setup(
    name=cfg["lib_name"],
    license='GPLv3+',
    version=cfg["version"],
    classifiers=[
        "Development Status :: " + statuses[int(cfg["status"])],
        "Intended Audience :: " + cfg["audience"].title(),
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: " + cfg["language"].title(),
    ]
    + [f"Programming Language :: Python :: {v}" for v in comp_version],
    url=f"https://github.com/{cfg['user']}/{cfg['lib_name']}",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    python_requires=">=" + cfg["min_python"],
    long_description=Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    project_urls={
        "Documentation": f"https://github.com/{cfg['user']}/{cfg['lib_name']}",
        "Tracker": f"https://github.com/{cfg['user']}/{cfg['lib_name']}/issues",
    },
)